import logging
import os
import re
from typing import Any, Mapping

import yaml

# Matches "${myvar-default}" -> "${", "myvar", "-", "default", "}"
# or "${myvar}" -> "${", "myvar", "", "", "}"
from configue.exceptions import NonCallableError

ENV_PATTERN_REGEX = re.compile(r"(?:(?!\${[^-}]+}).)*(\${)([^-}]+)(-?)([^}]*)(})")
CONSTRUCTOR_KEY = "()"
ESCAPED_CONSTRUCTOR_KEY = "\\()"


# pylint: disable=too-many-ancestors
class ConfigueLoader(yaml.FullLoader):
    logger = logging.getLogger(__name__)

    def construct_yaml_map(self, node: yaml.MappingNode) -> Mapping:
        mapping = self.construct_mapping(node)
        if isinstance(mapping, dict) and CONSTRUCTOR_KEY in mapping:
            cls = self.find_python_name(mapping.pop(CONSTRUCTOR_KEY), node.start_mark, unsafe=True)
            if not callable(cls):
                raise NonCallableError(
                    f"Error while constructing a Python instance {node.start_mark}, "
                    f"expected a callable but found {type(cls)}"
                )
            return cls(**mapping)
        if isinstance(mapping, dict) and ESCAPED_CONSTRUCTOR_KEY in mapping:
            mapping[CONSTRUCTOR_KEY] = mapping.pop(ESCAPED_CONSTRUCTOR_KEY)
        return mapping

    def construct_scalar(self, node: yaml.ScalarNode) -> Any:
        scalar = yaml.FullLoader.construct_scalar(self, node)
        replaced_value = ""
        end_pos = 0
        for match in ENV_PATTERN_REGEX.finditer(scalar):
            env_var_name, has_default, default_value = match.group(2, 3, 4)
            start_pos = match.start(1)
            if env_var_name not in os.environ and not has_default:
                self.logger.warning(f"Missing environment var: '{env_var_name}', no default is set")
            replaced_value += f"{scalar[end_pos:start_pos]}{os.environ.get(env_var_name, default_value)}"
            end_pos = match.end(5)
        replaced_value += scalar[end_pos:]
        if replaced_value == node.value:
            return scalar
        if node.style in ["'", '"']:
            replaced_value = f"{node.style}{replaced_value}{node.style}"
        # A variable has been replaced, reload to convert string to number if needed or replace again
        return yaml.load(replaced_value, Loader=self.__class__)



from typing import Any, List, Union

from .root_loader import RootLoader


def load(file_path: str, sub_path: Union[str, List[str]] = "", *, logging_config_path: str = None) -> Any:
    """Load configuration from a YAML file.

    :param file_path: Absolute path to the YAML file containing the configuration
    :param sub_path: path inside the YAML file to load. List elements are noted as 0-based indexes.
    :param logging_config_path: path inside the YAML file that contains the logging configuration.
    The format is the same as logging.dictConfig():
    https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
    :return: the converting dict corresponding to the file.

    Taking this file as an example:

    top_level_key:
      first_level: some_value
      other_key: 123
      some.dotted.key: dotted.value
      some_list:
        - item_key: item_value
          other_item_key: false
        - second_key: ~


    Loading the default sub_path (empty string) will return the whole object
    {
        "top_level_key": {
            "first_level": "some_value",
            "other_key": 123,
            "some.dotted.key": "dotted.value",
            "some_list": [{
                "item_key": "item_value",
                "other_item_key": False,
            }, {
                "second_key": None,
            }]
        }
    }

    Loading the sub_path "top_level_key.first_level" will return "some_value".
    Loading the sub_path "top_level_key.some_list.0" will return
    {
        "item_key": "item_value",
        "other_item_key": False,
    }
    Loading the sub_path ["top_level_key", "some.dotted.key"] will return "dotted.value"
    """

    return RootLoader(file_path).load_root_file(sub_path, logging_config_path)



import logging
import logging.config
from typing import Any, Dict, List, Optional, Union

from .file_loader import FileLoader


class RootLoader:
    logger = logging.getLogger(__name__)

    def __init__(self, file_path: str) -> None:
        self._root_file = file_path
        self._file_loaders_by_file: Dict[str, FileLoader] = {}

    def load_root_file(self, sub_path: Union[str, List[str]], logging_config_path: Optional[str]) -> Any:
        if logging_config_path is not None:
            self._load_logging_config(logging_config_path)
        return self.load_file(self._root_file, sub_path)

    def _load_logging_config(self, logging_config_path: str) -> None:
        logging.captureWarnings(True)
        logging_config = self.load_file(self._root_file, logging_config_path)
        logging.config.dictConfig(logging_config)

    def load_file(self, file_path: str, sub_path: Union[str, List[str]]) -> Any:
        if file_path not in self._file_loaders_by_file:
            self._file_loaders_by_file[file_path] = FileLoader(file_path, self)
        return self._file_loaders_by_file[file_path].load(sub_path)



class ConfigueError(Exception):
    pass


class SubPathNotFound(ConfigueError):
    pass


class NonCallableError(ConfigueError):
    pass


class ConfigueRecursionError(ConfigueError):
    pass


class NotFoundError(ConfigueError):
    pass


class InvalidNodeType(Exception):
    pass



import os

import setuptools

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="configue",
    version="4.2.0",
    url="https://github.com/illuin-tech/configue/",
    author="Illuin Technology",
    author_email="contact@illuin.tech",
    description="Helpers to load your application configuration from YAML files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    platforms="any",
    install_requires=[
        "pyyaml>=5.1.0,<7.0.0",
    ],
    python_requires=">=3.6,<4.0",
    packages=setuptools.find_packages(include=["configue", "configue.*"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
